unit ZLibClass;

{  -------------------------------------------------
   jkfStream 1.0

   Copyright (c) 2001 Kim Friis, Friisware.com
   kim@friisware.com

   This software is Freeware and without any
   warranty what so ever. Don't blame me if it
   crashes your computer, burns down your house or
   steals all your friends :-). (You get the picture)
   Please use at own risk.
   ------------------------------------------------}

interface

uses Classes, SysUtils, ZLib;

const
  BufferSize = 4096;

type
  TZLibProgressEvent = procedure(const Pct,CompRate:Double;const Position,Length:Integer) of object;

  TZLibPackType = (zpNone,zpFastest,zpMedium,zpMaximum);

  TZLibClass = class
  private
    ZStream: TCustomZLibStream;
    InStreamLength:Integer;
    Buffer: array[0..BufferSize-1] of Byte;
    FProgressEvent:TZLibProgressEvent;
    fPackType: TZLibPackType;
    procedure Status(Sender: TObject);
  public
    constructor Create;
    procedure Pack(InStream,OutStream:TStream);
    procedure UnPack(InStream,OutStream:TStream);
    property ProgressEvent:TZLibProgressEvent read FProgressEvent write FProgressEvent;
    property PackType:TZLibPackType read fPackType write fPackType default zpMedium;
  end;


implementation

procedure TZLibClass.Pack(InStream,OutStream:TStream);
begin
  if Assigned(InStream) and Assigned(OutStream) then begin
    InStreamLength := InStream.Size;
    ZStream := TCompressionStream.Create(TCompressionLevel(Ord(fPackType)), OutStream);
    try
      TCompressionStream(ZStream).OnProgress := Status;
      ZStream.CopyFrom(InStream, 0);
    finally
      Status(ZStream);
      ZStream.Free;
    end;
  end;
end;


procedure TZLibClass.UnPack(InStream,OutStream:TStream);
var
  Count: Integer;
begin
  if Assigned(InStream) and Assigned(OutStream) then begin
    InStreamLength := InStream.Size;
    InStream.Position := 0;
    ZStream := TDecompressionStream.Create(InStream);
    try
      TDecompressionStream(ZStream).OnProgress := Status;
      while True do
        begin
          Count := ZStream.Read(Buffer, BufferSize);
          if Count <> 0 then OutStream.WriteBuffer(Buffer, Count) else Break;
        end;
    finally
      Status(ZStream);
      ZStream.Free;
    end;
  end;
end;

procedure TZLibClass.Status(Sender: TObject);
var
  Pos,Length:Integer;
  CompRate:Single;
begin
  if Assigned(FProgressEvent) then begin
    Pos := TCustomZLibStream(Sender).Position;
    Length := InStreamLength;
    CompRate := 0;
    if (Sender is TCompressionStream) then
      CompRate := TCompressionStream(Sender).CompressionRate;
    FProgressEvent((Pos/Length)*100,CompRate,Pos,Length);
  end;
end;

constructor TZLibClass.Create;
begin
  inherited Create;
  fPackType := zpMedium;
end;

end.
