unit jkfStreams;

{  -------------------------------------------------
   jkfStream 1.0

   Copyright (c) 2001 Kim Friis, Friisware.com
   kim@friisware.com

   This software is Freeware and without any
   warranty what so ever. Don't blame me if it
   crashes your computer, burns down your house or
   steals all your friends :-). (You get the picture)
   Please use at own risk.
   ------------------------------------------------}

interface

uses Classes,sysutils,ZLibClass;

const
  JKFId2 = 'JKF';
  JKFVersion2 = '002.0';

type
  EJkfStreamException = class(Exception);

  TJKFPackHeader2 = record
    ID:Array[0..2] of Char;
    Version:Array[0..4] of Char;
    FilePacked:Integer;
    FileListSize:Integer;
  end;

  TJKFFileHeader = class(TObject)
    Name:String;
    SizeUnpacked:Int64;
    SizePacked:Int64;
    CRCCheck:Int64;
  end;

  TJkfFileList = class(TStringList)
  private
    function GetSizePacked(Index: Integer): Int64;
    function GetSizeUnpacked(Index: Integer): Int64;
  public
    destructor Destroy; override;
    procedure Delete(Index: Integer); override;
    procedure Clear; override;
    property SizesUnPacked[Index:Integer]:Int64 read GetSizeUnpacked;
    property SizesPacked[Index:Integer]:Int64 read GetSizePacked;
  end;

  TJkfStream = class
  private
    fPackedFile:TMemoryStream;
    fLoaded:Boolean;
    fHeader:TJKFPackHeader2;
    fPackType: TZLibPackType;
    procedure InternalView(Files:TJkfFileList);
    procedure InternalExtractHeader(var aHeader:TJKFPackHeader2);
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromFile(aFilename:String);
    procedure LoadFromStream(aStream:TStream);
    procedure SaveToFile(aFilename:String);
    procedure SaveToStream(aStream:TStream);
    procedure ExtractFileToStream(aFileName:String;aStream:TStream);
    procedure ExtractFileToFile(aFileName,DestFileName:String);
    procedure AddStream(aFileName:String;aStream:TStream);
    procedure AddFile(aFileName:String);
    procedure RemoveFile(aFileName:String);

    procedure Clear;
    procedure View(Files:TJkfFileList);

    property Loaded:Boolean read fLoaded;
    property PackType:TZLibPackType read fPackType write fPackType;
  end;

  TJkfPropertyStream = class(TPersistent)
  private
    fJkfStream:TJkfStream;
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadData(Stream: TStream); virtual;
    procedure WriteData(Stream: TStream); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure Assign(Source: TPersistent); override;
    procedure LoadFromStream(aStream:TStream);
    procedure SaveToStream(aStream:TStream);
    procedure LoadFromFile(const FileName : string);
    procedure SaveToFile(const FileName : string);
  end;

  TJkfStoredPack = class(TComponent)
  private
    fData:TJkfPropertyStream;
    function GetPackedFile: TJkfPropertyStream;
    procedure SetPackedFile(const Value: TJkfPropertyStream);
    function GetPackType: TZLibPackType;
    procedure SetPackType(const Value: TZLibPackType);
  public
    constructor Create(aOwner:TComponent); override;
    destructor Destroy; override;
    { procedures from the class }
    procedure LoadFromFile(aFilename:String);
    procedure LoadFromStream(aStream:TStream);
    procedure SaveToFile(aFilename:String);
    procedure SaveToStream(aStream:TStream);
    procedure ExtractFileToStream(aFileName:String;aStream:TStream);
    procedure ExtractFileToFile(aFileName,DestFileName:String);
    procedure AddStream(aFileName:String;aStream:TStream);
    procedure AddFile(aFileName:String);
    procedure RemoveFile(aFileName:String);
    procedure Clear;
    procedure View(Files:TJkfFileList);
  published
    property PackedFile:TJkfPropertyStream read GetPackedFile write SetPackedFile;
    property PackType:TZLibPackType read GetPackType write SetPackType default zpMedium;
  end;

implementation

{ TJkfStream }

procedure TJkfStream.AddFile(aFileName: String);
var
  FS:TFileStream;
begin
  if FileExists(aFileName) then begin
    FS := TFileStream.Create(aFileName,fmOpenRead);
    try
      AddStream(ExtractFileName(aFileName),FS);
    finally
      FS.Free;
    end;
  end else begin
  end;
end;

procedure TJkfStream.AddStream(aFileName: String; aStream: TStream);
var
  PackedStream,TmpPackedData,FileListHeader:TMemoryStream;
  SL:TJkfFileList;
  PackClass:TZLibClass;
  I,J:Integer;
  FileHeader:TJKFFileHeader;
  S2:String;
  S:Array[0..255] of Char;
begin
  if Assigned(aStream) then begin
    PackedStream := TMemoryStream.Create;
    try
      PackClass := TZLibClass.Create;
      try
        PackClass.PackType := PackType;
        PackClass.Pack(aStream,PackedStream);
      finally
        PackClass.Free;
      end;
      TmpPackedData := TMemoryStream.Create;
      try
        FileListHeader := TMemoryStream.Create;
        try
          SL := TJkfFileList.Create;
          try
            View(SL);
            for I := 0 to SL.Count - 1 do begin
              FileHeader := TJKFFileHeader(SL.Objects[I]);
              FileListHeader.Write(FileHeader.SizeUnpacked,SizeOf(FileHeader.SizeUnpacked));
              FileListHeader.Write(FileHeader.Sizepacked,SizeOf(FileHeader.Sizepacked));
              FileListHeader.Write(FileHeader.CRCCheck,SizeOf(FileHeader.CRCCheck));
              J := Length(SL[I]);
              FileListHeader.Write(J,SizeOf(Integer));
              S2 := SL[I];
              for J := 0 to Length(S2) do
                S[J] := S2[J+1];
              FileListHeader.Write(S,Length(S2));
            end;
            FileHeader := TJKFFileHeader.Create;
            try
              FileHeader.SizeUnpacked := aStream.Size;
              FileHeader.SizePacked := PackedStream.Size;
              FileHeader.CRCCheck := 0;
              FileListHeader.Write(FileHeader.SizeUnpacked,SizeOf(FileHeader.SizeUnpacked));
              FileListHeader.Write(FileHeader.Sizepacked,SizeOf(FileHeader.Sizepacked));
              FileListHeader.Write(FileHeader.CRCCheck,SizeOf(FileHeader.CRCCheck));
            finally
              FileHeader.Free;
            end;
            J := Length(aFileName);
            FileListHeader.Write(J,SizeOf(Integer));
            S2 := aFileName;
            for J := 0 to Length(S2) do
              S[J] := S2[J+1];
            FileListHeader.Write(S,Length(S2));
          finally
            SL.Free;
          end;
          fPackedFile.Position := SizeOf(fHeader)+fHeader.FileListSize;
          fHeader.FileListSize := FileListHeader.Size;
          if fPackedFile.Size > 0 then
            TmpPackedData.CopyFrom(fPackedFile,fPackedFile.Size-fPackedFile.Position);
          fPackedFile.Clear;
          fPackedFile.Write(fHeader.ID,SizeOf(fHeader.ID));
          fPackedFile.Write(fHeader.Version,SizeOf(fHeader.Version));
          fPackedFile.Write(fHeader.FilePacked,SizeOf(fHeader.FilePacked));
          fPackedFile.Write(fHeader.FileListSize,SizeOf(fHeader.FileListSize));
          fPackedFile.CopyFrom(FileListHeader,0);
          fPackedFile.CopyFrom(TmpPackedData,0);
          fPackedFile.CopyFrom(PackedStream,0);
          fLoaded := True;
        finally
          FileListHeader.Free;
        end;
      finally
        TmpPackedData.Free;
      end;
    finally
      PackedStream.Free;
    end;
  end;
end;

procedure TJkfStream.Clear;
begin
  fPackedFile.Clear;
end;

constructor TJkfStream.Create;
begin
  fPackType := zpMedium;
  fPackedFile := TMemoryStream.Create;
  fLoaded := False;
  fHeader.ID := JKFID2;
  fHeader.Version := JKFVersion2;
  fHeader.FilePacked := 2;
  fHeader.FileListSize := 0;
end;

destructor TJkfStream.Destroy;
begin
  fPackedFile.Free;
  inherited;
end;

procedure TJkfStream.ExtractFileToFile(aFileName, DestFileName: String);
var
  FS:TFileStream;
begin
  if Loaded then begin
    FS := TFileStream.Create(DestFileName,fmCreate);
    try
      ExtractFileToStream(aFileName,FS);
    finally
      FS.Free;
    end;
  end;
end;

procedure TJkfStream.ExtractFileToStream(aFileName: String;
  aStream: TStream);
var
  Pos,Offset,Size,I:Integer;
  FileList:TJkfFileList;
  PackedFile:TMemoryStream;
  PackClass:TZLibClass;
  FileHeader:TJKFFileHeader;
begin
  if Loaded then begin
    if Assigned(aStream) then begin
      FileList := TJkfFileList.Create;
      try
        InternalView(FileList);
        Pos := FileList.IndexOf(UpperCase(aFileName));
        if Pos > - 1 then begin
          Offset := 0;
          for I := 0 to Pos-1 do begin
            FileHeader := TJKFFileHeader(FileList.Objects[I]);
            Offset := Offset + FileHeader.SizePacked;
          end;
          FileHeader := TJKFFileHeader(FileList.Objects[Pos]);
          Size := FileHeader.SizePacked;
          fPackedFile.Seek(SizeOf(fHeader)+fHeader.FileListSize+OffSet,soFromBeginning);
          PackedFile := TMemoryStream.Create;
          try
            PackedFile.CopyFrom(fPackedFile,Size);
            PackClass := TZLibClass.Create;
            try
              PackClass.UnPack(PackedFile,aStream);
            finally
              PackClass.Free;
            end;
          finally
            PackedFile.Free;
          end;
        end;
      finally
        FileList.Free;
      end;
    end;
  end;
  aStream.Position := 0;
end;

procedure TJkfStream.InternalExtractHeader(var aHeader: TJKFPackHeader2);
begin
  if Loaded then begin
    fPackedFile.Position := 0;
    fPackedFile.Read(aHeader.ID ,SizeOf(aHeader.ID));
    fPackedFile.Read(aHeader.Version,SizeOf(aHeader.Version));
    fPackedFile.Read(aHeader.FilePacked,SizeOf(aHeader.FilePacked));
    fPackedFile.Read(aHeader.FileListSize,SizeOf(aHeader.FileListSize));
  end;
end;

procedure TJkfStream.InternalView(Files: TJkfFileList);
var
  I:Integer;
  Stor:Int64;
  S:Array[0..255] of Char;
  FileHeader:TJKFFileHeader;
begin
  if not Assigned(Files) then
    Files := TJkfFileList.Create;
  if Loaded then begin
    fPackedFile.Position := SizeOf(fHeader);
    while fPackedFile.Position<(fHeader.FileListSize+SizeOf(fHeader.FileListSize)) do begin
      FileHeader := TJKFFileHeader.Create;
      fPackedFile.Read(Stor,SizeOf(Stor));
      FileHeader.SizeUnpacked := Stor;
      fPackedFile.Read(Stor,SizeOf(Stor));
      FileHeader.Sizepacked := Stor;
      fPackedFile.Read(Stor,SizeOf(Stor));
      FileHeader.CRCCheck := Stor;
      fPackedFile.Read(I,SizeOf(Integer));
      fPackedFile.Read(S,I);
      S[I] := #0;
      FileHeader.Name := UpperCase(S);
      Files.AddObject(S,FileHeader);
    end;
  end;
end;

procedure TJkfStream.LoadFromFile(aFilename: String);
var
  FS:TFileStream;
begin
  if FileExists(aFileName) then begin
    FS := TFileStream.Create(aFileName,fmOpenRead);
    try
      LoadFromStream(FS);
    finally
      FS.Free;
    end;
  end else begin
    //Nothing to load
    fPackedFile.Clear;
  end;
end;

procedure TJkfStream.LoadFromStream(aStream: TStream);
begin
  if Assigned(aStream) then begin
    fPackedFile.LoadFromStream(aStream);
    fLoaded := True;
    InternalExtractHeader(fHeader);
    if (fHeader.ID <> JKFId2) or (fHeader.Version <> JKFVersion2) then begin
      fPackedFile.Clear;
      fLoaded := False;
      raise EJkfStreamException.Create('Error reading file, not proper file format');
    end;
  end else begin
    fPackedFile.Clear;
  end;
end;

procedure TJkfStream.RemoveFile(aFileName: String);
var
  Pos,Offset,Size,I,J:Integer;
  FileList:TJkfFileList;
  PackedFile,FileListHeader:TMemoryStream;
  FileHeader:TJKFFileHeader;
  S:Array[0..255] of Char;
  S2:String;
begin
  if Loaded then begin
    FileList := TJkfFileList.Create;
    try
      InternalView(FileList);
      Pos := FileList.IndexOf(UpperCase(aFileName));
      if Pos > - 1 then begin
        Offset := 0;
        FileHeader := TJKFFileHeader.Create;
        try
          for I := 0 to Pos-1 do begin
            FileHeader := TJKFFileHeader(FileList.Objects[I]);
            Offset := Offset + FileHeader.SizePacked;
          end;
          FileHeader := TJKFFileHeader(FileList.Objects[Pos]);
          Size := FileHeader.SizePacked;
        finally
          FileHeader.Free;
        end;
        fPackedFile.Seek(SizeOf(fHeader)+fHeader.FileListSize,soFromBeginning);
        PackedFile := TMemoryStream.Create;
        try
          PackedFile.CopyFrom(fPackedFile,OffSet);
          fPackedFile.Seek(Size,soFromCurrent);
          PackedFile.CopyFrom(fPackedFile,fPackedFile.Size-(OffSet+Size+SizeOf(fHeader)+fHeader.FileListSize));
          fPackedFile.Clear;
          FileListHeader := TMemoryStream.Create;
          try
            FileList.Delete(Pos);
            for I := 0 to FileList.Count - 1 do begin
              FileHeader := TJKFFileHeader(FileList.Objects[I]);
              FileListHeader.Write(FileHeader.SizeUnpacked,SizeOf(FileHeader.SizeUnpacked));
              FileListHeader.Write(FileHeader.Sizepacked,SizeOf(FileHeader.Sizepacked));
              FileListHeader.Write(FileHeader.CRCCheck,SizeOf(FileHeader.CRCCheck));
              J := Length(FileList[I]);
              FileListHeader.Write(J,SizeOf(Integer));
              S2 := FileList[I];
              for J := 0 to Length(S2) do
                S[J] := S2[J+1];
              FileListHeader.Write(S,Length(S2));
            end;
            fHeader.FileListSize := FileListHeader.Size;
            fPackedFile.Clear;
            fPackedFile.Write(fHeader.ID,SizeOf(fHeader.ID));
            fPackedFile.Write(fHeader.Version,SizeOf(fHeader.Version));
            fPackedFile.Write(fHeader.FilePacked,SizeOf(fHeader.FilePacked));
            fPackedFile.Write(fHeader.FileListSize,SizeOf(fHeader.FileListSize));
            fPackedFile.CopyFrom(FileListHeader,0);
            fPackedFile.CopyFrom(PackedFile,0);
          finally
            FileListHeader.Free;
          end;
        finally
          PackedFile.Free;
        end;
      end;
    finally
      FileList.Free;
    end;
  end;
end;

procedure TJkfStream.SaveToFile(aFilename: String);
begin
  fPackedFile.SaveToFile(aFileName);
end;

procedure TJkfStream.SaveToStream(aStream: TStream);
begin
  if Assigned(aStream) then
    fPackedFile.SaveToStream(aStream);
end;

procedure TJkfStream.View(Files: TJkfFileList);
begin
  InternalView(Files);
end;

{ TJkfStoredPack }

procedure TJkfStoredPack.AddFile(aFileName: String);
begin
  fData.fJkfStream.AddFile(aFileName);
end;

procedure TJkfStoredPack.AddStream(aFileName: String; aStream: TStream);
begin
  fData.fJkfStream.AddStream(aFileName,aStream);
end;

procedure TJkfStoredPack.Clear;
begin
  fData.fJkfStream.Clear;
end;

constructor TJkfStoredPack.Create(aOwner: TComponent);
begin
  inherited;
  fData := TJkfPropertyStream.Create;
end;

destructor TJkfStoredPack.Destroy;
begin
  fData.Free;
  inherited;
end;

procedure TJkfStoredPack.ExtractFileToFile(aFileName,
  DestFileName: String);
begin
  fData.fJkfStream.ExtractFileToFile(aFileName,DestFileName);
end;

procedure TJkfStoredPack.ExtractFileToStream(aFileName: String;
  aStream: TStream);
begin
  fData.fJkfStream.ExtractFileToStream(aFileName,aStream);
end;

function TJkfStoredPack.GetPackedFile: TJkfPropertyStream;
begin
  Result := fData;
end;

function TJkfStoredPack.GetPackType: TZLibPackType;
begin
  result := fData.fJkfStream.PackType;
end;

procedure TJkfStoredPack.LoadFromFile(aFilename: String);
begin
  fData.fJkfStream.LoadFromFile(aFilename);
end;

procedure TJkfStoredPack.LoadFromStream(aStream: TStream);
begin
  fData.fJkfStream.LoadFromStream(aStream);
end;

procedure TJkfStoredPack.RemoveFile(aFileName: String);
begin
  fData.fJkfStream.RemoveFile(aFileName);
end;

procedure TJkfStoredPack.SaveToFile(aFilename: String);
begin
  fData.fJkfStream.SaveToFile(aFilename);
end;

procedure TJkfStoredPack.SaveToStream(aStream: TStream);
begin
  fData.fJkfStream.SaveToStream(aStream);
end;

procedure TJkfStoredPack.SetPackedFile(const Value: TJkfPropertyStream);
begin
  fData := Value;
end;

procedure TJkfStoredPack.SetPackType(const Value: TZLibPackType);
begin
  fData.fJkfStream.PackType := Value;
end;

procedure TJkfStoredPack.View(Files: TJkfFileList);
begin
  fData.fJkfStream.View(Files);
end;

{ TJkfPropertyStream }

procedure TJkfPropertyStream.Assign(Source: TPersistent);
begin
  if Source=nil then
  begin
    Clear;
  end else if Source is TJkfPropertyStream then begin
    fJkfStream.LoadFromStream(TJkfPropertyStream(Source).fJkfStream.fPackedFile);
  end else
    inherited Assign(Source);
end;

procedure TJkfPropertyStream.Clear;
begin
  fJkfStream.Clear;
end;

constructor TJkfPropertyStream.Create;
begin
  fJkfStream := TJkfStream.Create;
end;

procedure TJkfPropertyStream.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('PAKDATA', ReadData, WriteData, True);
end;

destructor TJkfPropertyStream.Destroy;
begin
  fJkfStream.Free;
  inherited;
end;

procedure TJkfPropertyStream.LoadFromFile(const FileName: string);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TJkfPropertyStream.LoadFromStream(aStream: TStream);
begin
  fJkfStream.LoadFromStream(aStream);
end;

procedure TJkfPropertyStream.ReadData(Stream: TStream);
begin
  LoadFromStream(Stream);
end;

procedure TJkfPropertyStream.SaveToFile(const FileName: string);
var
  Stream: TFileStream;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    SaveToStream(Stream);
  finally
    Stream.Free;
  end;
end;

procedure TJkfPropertyStream.SaveToStream(aStream: TStream);
begin
  fJkfStream.SaveToStream(aStream);
end;

procedure TJkfPropertyStream.WriteData(Stream: TStream);
begin
  SaveToStream(Stream);
end;

{ TJkfFileList }

procedure TJkfFileList.Clear;
var
  I:Integer;
begin
  for I := 0 to Count - 1 do begin
    TJKFFileHeader(Objects[I]).Free;
  end;
  inherited;
end;

procedure TJkfFileList.Delete(Index: Integer);
begin
  TJKFFileHeader(Objects[Index]).Free;
  inherited;
end;

destructor TJkfFileList.Destroy;
begin
  Clear;
  inherited;
end;

function TJkfFileList.GetSizePacked(Index: Integer): Int64;
begin
  Result := -1;
  if (Index < Count) and (Index > -1) then begin
    Result := TJKFFileHeader(Objects[Index]).SizePacked;
  end;
end;

function TJkfFileList.GetSizeUnpacked(Index: Integer): Int64;
begin
  Result := -1;
  if (Index < Count) and (Index > -1) then begin
    Result := TJKFFileHeader(Objects[Index]).SizeUnPacked;
  end;
end;

end.
